#define _BSD_SOURCE
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/route.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>

//#include "simplecom.h"
//#include "dslconf.h"
#include "stealip.h"

#define dprint(...) \
	fprintf(stderr, "%s(): ", __func__); \
	fprintf(stderr, __VA_ARGS__); \
	fprintf(stderr, "\n");
//#define dnotify dprint
#define dnotify(...) /* nothing */
#define dwarn dprint
#define derror dprint

static in_addr_t addr = 0, peer = 0;
static char routes_applied = 0;
static struct rtentry routes[3];

#define DISABLE_FILE "/tmp/stealproof"
#define LAN_IFNAME "br0"	/* LAN is bridged */

static const char *wan_ifnames[] = {
	"ppp0",		/* for PPP connections */
#ifdef TARGET_FUSIV
	"atm0",		/* IPoA */
#endif
	NULL		/* terminating entry */
};

enum iface_state {
	IFS_NOT_EXISTING = 0,
	IFS_NO_IP_ASSIGNED = 1,
	IFS_WITH_IP = 2,
};

static int do_ioctl(int sockfd, unsigned long type, void *data)
{
	if (ioctl(sockfd, type, data)) {
		perror("ioctl()");
		return 1;
	}
	return 0;
}

/* return true if writing to fd failed */
static int __write_ip_out(int fd, in_addr_t addr)
{
	struct in_addr addr_s = {
		.s_addr = addr,
	};
	char *p = inet_ntoa(addr_s);

	return !(write(fd, p, strlen(p)) == (signed)strlen(p));
}

/* write to PUB_IP_FILE:
 * <addr>/<pfx> <gw>
 * (where pfx is always 32 and gw the ppp peer)
 */
static int write_ip_out(void)
{
	int fd;

	dnotify("being called");

	if ((fd = open(PUB_IP_FILE, O_TRUNC | O_CREAT | O_WRONLY, S_IRUSR | S_IWUSR)) == -1) {
		perror("open()");
		return 1;
	}

	if (!addr || !peer) {
		write(fd, "none", 4);
		close(fd);
		return 0;
	}

	dnotify("setting IPs to %d and %d", addr, peer);

	if (__write_ip_out(fd, addr) ||
	    write(fd, "/32 ", 4) != 4 ||
	    __write_ip_out(fd, peer)) {
		perror("write()");
		close(fd);
		return 1;
	}
	close(fd);
	return 0;
}

/* reset the recorded IPs to zero, return true if they weren't before */
static int reset_ips(void)
{
	if (addr || peer) {
		addr = peer = 0;
		return 1;
	}
	return 0;
}

static enum iface_state fetch_ip(int sockfd, const char *ifname)
{
	struct ifreq ifr;
	struct rtentry *rtp;
	enum iface_state ret = IFS_NO_IP_ASSIGNED;

	if (!if_nametoindex(ifname))	/* interface doesn't exist */
		return IFS_NOT_EXISTING;

	/* following ioctl may fail (no IP assigned) */
	memset(&ifr, 0, sizeof(ifr));
	snprintf(ifr.ifr_name, IFNAMSIZ, "%s", ifname);
	ioctl(sockfd, SIOCGIFADDR, &ifr);

	if (((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr) {
		dnotify("WAN address has changed");
		addr = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr;
		ret = IFS_WITH_IP;
	} else {
		dnotify("WAN address has not changed");
	}

	/* following ioctl may fail (no IP assigned) */
	memset(&ifr, 0, sizeof(ifr));
	snprintf(ifr.ifr_name, IFNAMSIZ, "%s", ifname);
	ioctl(sockfd, SIOCGIFDSTADDR, &ifr);

	if (((struct sockaddr_in *)&ifr.ifr_dstaddr)->sin_addr.s_addr) {
		dnotify("ppp-dst-address has changed");
		peer = ((struct sockaddr_in *)&ifr.ifr_dstaddr)->sin_addr.s_addr;
		ret = IFS_WITH_IP;
	} else {
		dnotify("ppp-dst-address has not changed");
	}

	if (ret == IFS_NO_IP_ASSIGNED)
		return ret;

	/* remove the address from the interface */
	memset(&ifr, 0, sizeof(ifr));
	snprintf(ifr.ifr_name, IFNAMSIZ, "%s", ifname);
	((struct sockaddr_in *)&ifr.ifr_addr)->sin_family = AF_INET;
	do_ioctl(sockfd, SIOCSIFADDR, &ifr);
	dnotify("removed addresses, please check");

	/* add the necessary routes */

	if (routes_applied) {
		ioctl(sockfd, SIOCDELRT, &routes[0]);
		ioctl(sockfd, SIOCDELRT, &routes[1]);
		ioctl(sockfd, SIOCDELRT, &routes[2]);
	}
	memset(routes, 0, 3 * sizeof(struct rtentry));
	routes_applied = 0;

	/* add the ppp peer hostroute */
	dnotify("adding ppp dst hostroute");
	rtp = &routes[0];
	((struct sockaddr_in *)&rtp->rt_dst)->sin_family = AF_INET;
	((struct sockaddr_in *)&rtp->rt_dst)->sin_addr.s_addr = peer;
	rtp->rt_flags = RTF_UP | RTF_HOST;
	/* reuse ifr.ifr_name here, as ifname is const */
	rtp->rt_dev = ifr.ifr_name;
	do_ioctl(sockfd, SIOCADDRT, rtp);

	/* add the default route */
	dnotify("adding default route");
	rtp = &routes[1];
	((struct sockaddr_in *)&rtp->rt_dst)->sin_family = AF_INET;
	((struct sockaddr_in *)&rtp->rt_gateway)->sin_family = AF_INET;
	((struct sockaddr_in *)&rtp->rt_genmask)->sin_family = AF_INET;
	((struct sockaddr_in *)&rtp->rt_dst)->sin_addr.s_addr = 0;
	((struct sockaddr_in *)&rtp->rt_gateway)->sin_addr.s_addr = peer;
	((struct sockaddr_in *)&rtp->rt_genmask)->sin_addr.s_addr = 0;
	rtp->rt_flags = RTF_UP | RTF_GATEWAY;
	do_ioctl(sockfd, SIOCADDRT, rtp);

	/* add the host route to viprinux */
	/* not needed on fusiv, as LAN_IFNAME is default route
	 * for anything from WAN interfaces via policy routing */
#ifndef TARGET_FUSIV
	dnotify("adding viprinux host route");
	rtp = &routes[2];
	((struct sockaddr_in *)&rtp->rt_dst)->sin_family = AF_INET;
	((struct sockaddr_in *)&rtp->rt_dst)->sin_addr.s_addr = addr;
	rtp->rt_flags = RTF_UP | RTF_HOST;
	rtp->rt_dev = LAN_IFNAME;
	do_ioctl(sockfd, SIOCADDRT, rtp);
#endif

	routes_applied = 1;
	return ret;
}

int main(void)
{
	int sockfd;

	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("socket()");
		return errno;
	}

	while (1) {
		int idx, rc;

		sleep(2);

		if (!access(DISABLE_FILE, F_OK))
			continue;

		for (idx = 0; wan_ifnames[idx]; idx++) {
			/* we are interested in two things:
			 * - does the interface exist?
			 *   -> this WAN interface is the used one
			 * - is an IP assigned?
			 *   -> IP has changed, needs to be stolen
			 */
			if ((rc = fetch_ip(sockfd, wan_ifnames[idx])))
				break;
		}

		if (wan_ifnames[idx]) {
			if (rc == IFS_WITH_IP)
				write_ip_out();
		} else {
			/* no iface exists, so no active WAN connection */
			reset_ips();
		}
	}

	return 0;
}
